import pytest

from izulu import _reraise


def test_fatal_direct_inheritance():
    type("Klass", (_reraise.FatalMixin,), {})


def test_fatal_indirect_inheritance():
    kls = type("Klass", (_reraise.FatalMixin,), {})

    with pytest.raises(TypeError):
        type("Klass2", (kls,), {})



import datetime
import types
from unittest import mock

import pytest

from izulu import _utils
from izulu import root
from tests import errors

TS = datetime.datetime.now(datetime.timezone.utc)


@mock.patch("izulu.root.Error._override_message")
@mock.patch("izulu.root.Error._Error__process_template")
@mock.patch("izulu.root.Error._Error__populate_attrs")
@mock.patch("izulu.root.Error._Error__process_toggles")
def test_init(
    fake_proc_ftrs,
    fake_set_attrs,
    fake_proc_tpl,
    fake_override_message,
):
    fake_proc_tpl.return_value = errors.RootError.__template__
    overriden_message = "overriden message"
    fake_override_message.return_value = overriden_message
    store = errors.RootError._Error__cls_store
    manager = mock.Mock()
    manager.attach_mock(fake_proc_ftrs, "fake_proc_ftrs")
    manager.attach_mock(fake_set_attrs, "fake_set_attrs")
    manager.attach_mock(fake_proc_tpl, "fake_proc_tpl")
    manager.attach_mock(fake_override_message, "fake_override_message")
    fake_override_message_call = mock.call.fake_override_message(
        store,
        {},
        errors.RootError.__template__,
    )
    expected_calls = [
        mock.call.fake_proc_ftrs(),
        mock.call.fake_set_attrs(),
        mock.call.fake_proc_tpl({}),
        fake_override_message_call,
    ]

    e = errors.RootError()

    assert manager.mock_calls == expected_calls
    assert str(e) == overriden_message


@pytest.mark.parametrize(
    ("kls", "kwargs", "msg", "attrs", "not_attrs"),
    [
        (errors.RootError, dict(), "Unspecified error", dict(), tuple()),
        (
            errors.TemplateOnlyError,
            dict(name="John", age=42),
            "The John is 42 years old",
            dict(),
            ("name", "age"),
        ),
        (
            errors.ComplexTemplateOnlyError,
            dict(name="John", age=42, ts=TS),
            (
                "********John********  42.000000"
                f" 0b101010 {TS:%Y-%m-%d %H:%M:%S}"
            ),
            dict(),
            ("name", "age", "ts"),
        ),
        (
            errors.AttributesOnlyError,
            dict(name="John", age=42),
            "Static message template",
            dict(name="John", age=42),
            tuple(),
        ),
        (
            errors.AttributesWithStaticDefaultsError,
            dict(name="John"),
            "Static message template",
            dict(name="John", age=0),
            tuple(),
        ),
        (
            errors.AttributesWithDynamicDefaultsError,
            dict(name="John"),
            "Static message template",
            dict(name="John", age=0),
            tuple(),
        ),
        (
            errors.ClassVarsError,
            dict(),
            "Static message template",
            dict(name="Username", age=42),
            tuple(),
        ),
        (
            errors.MixedError,
            dict(name="John", age=10, note="...", timestamp=TS),
            "The John is 10 years old with ...",
            dict(
                entity="The Entity",
                name="John",
                age=10,
                timestamp=TS,
                my_type="MixedError",
            ),
            ("note",),
        ),
        (
            errors.DerivedError,
            dict(
                name="John",
                age=10,
                note="...",
                timestamp=TS,
                surname="Brown",
                box=dict(a=11),
            ),
            "The John Brown is 10 years old with ...",
            dict(
                entity="The Entity",
                name="John",
                age=10,
                timestamp=TS,
                my_type="DerivedError",
                surname="Brown",
                box=dict(a=11),
            ),
            ("note",),
        ),
    ],
)
def test_instantiate_ok(kls, kwargs, msg, attrs, not_attrs):
    e = kls(**kwargs)

    assert str(e) == msg
    for attr, value in attrs.items():
        assert getattr(e, attr) == value
    for attr in not_attrs:
        assert not hasattr(e, attr)


@pytest.mark.parametrize(
    ("kls", "kwargs"),
    [
        (errors.RootError, dict(name="John")),
        (errors.TemplateOnlyError, dict(name="John")),
        (errors.AttributesOnlyError, dict(age=42)),
        (errors.AttributesWithStaticDefaultsError, dict()),
        (errors.AttributesWithDynamicDefaultsError, dict(age=0)),
        (errors.ClassVarsError, dict(name="John")),
    ],
)
def test_instantiate_fail(kls, kwargs):
    with pytest.raises(TypeError):
        kls(**kwargs)


@pytest.mark.parametrize(
    ("kls", "kwargs", "expected_kwargs"),
    [
        (errors.RootError, dict(), dict()),
        (errors.RootError, dict(name="John"), dict(name="John")),
        (errors.ClassVarsError, dict(), dict(name="Username", age=42)),
        (errors.ClassVarsError, dict(name="John"), dict(name="John", age=42)),
        (
            errors.MixedError,
            dict(name="John", surname="Brown", updated_at=TS, timestamp=TS),
            dict(
                name="John",
                surname="Brown",
                age=0,
                my_type="MixedError",
                entity="The Entity",
                updated_at=TS,
                timestamp=TS,
            ),
        ),
        (
            errors.MixedError,
            dict(updated_at=TS, timestamp=TS),
            dict(
                age=0,
                my_type="MixedError",
                entity="The Entity",
                updated_at=TS,
                timestamp=TS,
            ),
        ),
        (
            errors.DerivedError,
            dict(name="John", surname="Brown", updated_at=TS, timestamp=TS),
            dict(
                name="John",
                surname="Brown",
                full_name="John Brown",
                age=0,
                my_type="DerivedError",
                entity="The Entity",
                location=(50.3, 3.608),
                updated_at=TS,
                timestamp=TS,
            ),
        ),
    ],
)
@mock.patch("izulu._utils.format_template")
def test_process_template(mock_format, kls, kwargs, expected_kwargs):
    with mock.patch.object(
        kls,
        "__toggles__",
        new_callable=mock.PropertyMock,
    ) as mocked:
        mocked.return_value = root.Toggles.NONE
        kls(**kwargs)

    mock_format.assert_called_once_with(kls.__template__, expected_kwargs)


def test_override_message():
    e = errors.RootError()
    orig_msg = "my message"

    final_msg = e._override_message(e._Error__cls_store, dict(), orig_msg)

    assert final_msg == orig_msg
    assert id(final_msg) == id(orig_msg)


def test_constants():
    e = errors.RootError()
    toggles = root.Toggles.DEFAULT ^ root.Toggles.FORBID_UNANNOTATED_FIELDS

    assert e.__template__ == "Unspecified error"
    assert e.__toggles__ == toggles
    assert e._Error__cls_store == _utils.Store(
        fields=frozenset(),
        const_hints=types.MappingProxyType(dict()),
        inst_hints=types.MappingProxyType(dict()),
        consts=types.MappingProxyType(dict()),
        defaults=frozenset(),
    )



from __future__ import annotations

import _string  # type: ignore[import-not-found]  # noqa: PLC2701
import dataclasses
import string
import typing as t

if t.TYPE_CHECKING:
    import types

_IZULU_ATTRS = {
    "__template__",
    "__toggles__",
    "_Error__cls_store",
    "__reraising__",
    "_ReraisingMixin__reraising",
}
_FORMATTER = string.Formatter()


# TODO(d.burmistrov): dataclass options
@dataclasses.dataclass
class Store:
    fields: t.FrozenSet[str]
    const_hints: types.MappingProxyType[str, type]
    inst_hints: types.MappingProxyType[str, type]
    consts: types.MappingProxyType[str, t.Any]
    defaults: t.FrozenSet[str]

    registered: t.FrozenSet[str] = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        self.registered = self.fields.union(self.inst_hints)


def check_missing_fields(store: Store, kws: t.FrozenSet[str]) -> None:
    missing = store.registered.difference(store.defaults, store.consts, kws)
    if missing:
        raise TypeError(f"Missing arguments: {join_items(missing)}")


def check_undeclared_fields(store: Store, kws: t.FrozenSet[str]) -> None:
    undeclared = kws.difference(store.registered, store.const_hints)
    if undeclared:
        raise TypeError(f"Undeclared arguments: {join_items(undeclared)}")


def check_kwarg_consts(store: Store, kws: t.FrozenSet[str]) -> None:
    consts = kws.intersection(store.const_hints)
    if consts:
        raise TypeError(f"Constants in arguments: {join_items(consts)}")


def check_non_named_fields(store: Store) -> None:
    for field in store.fields:
        if isinstance(field, int):
            msg = f"Field names can't be digits: {field}"
            raise ValueError(msg)  # noqa: TRY004
        if not field:
            raise ValueError("Field names can't be empty")


def check_unannotated_fields(store: Store) -> None:
    unannotated = store.fields - set(store.const_hints) - set(store.inst_hints)
    if unannotated:
        msg = f"Fields must be annotated: {join_items(unannotated)}"
        raise ValueError(msg)


def join_items(items: t.Iterable[str]) -> str:
    return ", ".join(map("'{}'".format, items))


def join_kwargs(**kwargs: t.Any) -> str:  # noqa: ANN401
    return ", ".join(f"{k!s}={v!r}" for k, v in kwargs.items())


def format_template(template: str, kwargs: t.Dict[str, t.Any]) -> str:
    try:
        return template.format_map(kwargs)
    except Exception as e:
        msg_part = "Failed to format template with provided kwargs: "
        raise ValueError(msg_part + join_kwargs(**kwargs)) from e


def iter_fields(template: str) -> t.Generator[str, None, None]:
    # https://docs.python.org/3/library/string.html#format-string-syntax
    for _, fn, _, _ in _FORMATTER.parse(str(template)):
        if fn is not None:
            yield _string.formatter_field_name_split(fn)[0]


def split_cls_hints(
    cls: type,
) -> t.Tuple[t.Dict[str, type], t.Dict[str, type]]:
    const_hints: t.Dict[str, type] = {}
    inst_hints: t.Dict[str, type] = {}

    for k, v in t.get_type_hints(cls).items():
        if k in _IZULU_ATTRS:
            continue
        if t.get_origin(v) is t.ClassVar:
            const_hints[k] = v
        else:
            inst_hints[k] = v

    return const_hints, inst_hints


def get_cls_defaults(
    cls: type,
    attrs: t.Iterable[str],
) -> t.Dict[str, t.Any]:
    return {attr: getattr(cls, attr) for attr in attrs if hasattr(cls, attr)}


def traverse_tree(cls: type) -> t.Generator[type, None, None]:
    workload = cls.__subclasses__()
    discovered = []
    while workload:
        item = workload.pop()
        discovered.append(item)
        workload.extend(item.__subclasses__())
    yield from discovered



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from importlib import metadata

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "izulu"
author = "Dima Burmistrov"
copyright = "2023-%Y, " + author
# -- setuptools_scm ----------------------------------------------------------
version = metadata.version(project)
release = ".".join(version.split(".")[:3])


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.todo",           # Support for todo items
    "sphinx.ext.viewcode",       # Add links to highlighted source code
    "sphinx.ext.intersphinx",    # Link to other projects’ documentation
    # custom extentions
    "sphinx_copybutton",         # add a little “copy” button to the right of your code blocks
    "sphinx_design",             # for designing beautiful, screen-size responsive web-components
    "sphinx_favicon",
    "sphinx_togglebutton",
    # not used
    "sphinx.ext.autodoc",      # Include documentation from docstrings  # check sphinx.ext.apidoc
    "sphinx.ext.napoleon",     # Support for NumPy and Google style docstrings
    # "sphinx.ext.autosummary",  # Generate autodoc summaries
    # "sphinx.ext.graphviz",     # Add Graphviz graphs
]

autosummary_generate = True

exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_logo = "https://repository-images.githubusercontent.com/766241795/85494614-5974-4b26-bfec-03b8e393c7f0"

templates_path = ["_templates"]
html_static_path = ["_static"]
html_js_files = ["_js/custom-icon.js"]
html_sourcelink_suffix = ""

html_theme_options = {
    "use_edit_page_button": True,

    "show_toc_level": 2,

    "secondary_sidebar_items": {
      "**/*": ["page-toc", "edit-this-page", "sourcelink"],
      "*": ["page-toc", "edit-this-page", "sourcelink"],
      "index": [],
    },

    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/pyctrl/izulu",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://www.pypi.org",
            "icon": "fa-custom fa-pypi",
        },
        {
            "name": "pyctrl",
            "url": "https://github.com/pyctrl",
            "icon": "https://github.com/pyctrl/pyctrl/blob/main/logo/pyctrl/gray-460x460.png?raw=true",
            "type": "url",
        },
    ],

    "footer_start": ["copyright"],
    "footer_center": ["sphinx-version"],
}

html_context = {
    "github_url": "https://github.com",
    "github_user": "pyctrl",
    "github_repo": "izulu",
    "github_version": "main",
    "doc_path": "docs/source/",
}

favicons = [
   {
      "sizes": "16x16",
      "href": "https://github.com/pyctrl/pyctrl/blob/main/logo/izulu/izulu_logo_512.png?raw=true",
   },
   {
      "sizes": "32x32",
      "href": "https://github.com/pyctrl/pyctrl/blob/main/logo/izulu/izulu_logo_512.png?raw=true",
   },
]



from __future__ import annotations

import copy
import enum
import functools
import logging
import types
import typing as t

from izulu import _utils

_IMPORT_ERROR_TEXTS = (
    "",
    "You have early version of Python.",
    "  Extra compatibility dependency required.",
    "  Please add 'izulu[compatibility]' to your project dependencies.",
    "",
    "Pip: `pip install izulu[compatibility]`",
)


if hasattr(t, "dataclass_transform"):
    t_ext = t
else:
    try:
        import typing_extensions as t_ext  # type: ignore[no-redef]
    except ImportError:
        for message in _IMPORT_ERROR_TEXTS:
            logging.error(message)  # noqa: LOG015,TRY400
        raise

FactoryReturnType = t.TypeVar("FactoryReturnType")


class ErrorDumpDict(t.TypedDict):
    type: str
    reason: str
    fields: t.Dict[str, t.Any] | None
    details: t.Dict[t.Any, t.Any]


@t.overload
def factory(
    *,
    default_factory: t.Callable[[], FactoryReturnType],
    self: t.Literal[False] = False,
) -> FactoryReturnType: ...


@t.overload
def factory(
    *,
    default_factory: t.Callable[[Error], FactoryReturnType],
    self: t.Literal[True],
) -> FactoryReturnType: ...


def factory(
    *,
    default_factory: t.Callable[..., t.Any],
    self: bool = False,
) -> t.Any:
    """
    Attaches factory for dynamic default values.

    Args:
        default_factory: callable factory receiving 0 or 1 argument
            (see ``self`` param)
        self: controls callable factory argument
            if ``True`` factory will receive single argument of error instance
            otherwise factory will be invoked without argument

    """
    target = default_factory if self else (lambda _: default_factory())
    return functools.cached_property(target)


class Toggles(enum.Flag):
    FORBID_MISSING_FIELDS = enum.auto()
    FORBID_UNDECLARED_FIELDS = enum.auto()
    FORBID_KWARG_CONSTS = enum.auto()
    FORBID_NON_NAMED_FIELDS = enum.auto()
    FORBID_UNANNOTATED_FIELDS = enum.auto()

    NONE = 0
    DEFAULT = (
        FORBID_MISSING_FIELDS
        | FORBID_UNDECLARED_FIELDS
        | FORBID_KWARG_CONSTS
        | FORBID_NON_NAMED_FIELDS
        | FORBID_UNANNOTATED_FIELDS
    )


@t_ext.dataclass_transform(
    eq_default=False,
    order_default=False,
    kw_only_default=True,
    frozen_default=False,
    field_specifiers=(factory,),
)
class Error(Exception):
    """
    Base class for your exception trees.

    Example::

        class MyError(root.Error):
            __template__ = "{smth} has happened at {ts}"

            smth: str
            ts: root.factory(datetime.now)

    Provides 4 main features:

    * Instead of manual error message formatting (and copying it all over
      the codebase) provide just ``kwargs``:

      - before: ``raise MyError(f"{smth} has happened at {datetime.now()}")``
      - after: ``raise MyError(smth=smth)``

      Just provide ``__template__`` class attribute with your error message
      template string. New style formatting is used:

      - ``str.format()``
      - https://pyformat.info/
      - https://docs.python.org/3/library/string.html#formatspec

    * Automatic ``kwargs`` conversion into error instance attributes

    * You can attach static and dynamic default values:
      this is why ``datetime.now()`` was omitted above

    * Out-of-box validation for provided ``kwargs``
      (individually enable/disable checks with ``__toggles__`` attribute)

    """

    __template__: t.ClassVar[str] = "Unspecified error"
    __toggles__: t.ClassVar[Toggles] = Toggles.DEFAULT

    __cls_store: t.ClassVar[_utils.Store] = _utils.Store(
        fields=frozenset(),
        const_hints=types.MappingProxyType(dict()),
        inst_hints=types.MappingProxyType(dict()),
        consts=types.MappingProxyType(dict()),
        defaults=frozenset(),
    )

    def __iter__(self) -> t.Iterator[BaseException]:
        """Return iterator over the whole exception chain."""
        yield self
        cause = self.__cause__
        while cause is not None:
            yield cause
            exc = cause
            cause = exc.__cause__

    def __init_subclass__(cls, **kwargs: t.Any) -> None:  # noqa: ANN401
        super().__init_subclass__(**kwargs)
        fields = frozenset(_utils.iter_fields(cls.__template__))
        const_hints, inst_hints = _utils.split_cls_hints(cls)
        consts = _utils.get_cls_defaults(cls, const_hints)
        defaults = _utils.get_cls_defaults(cls, inst_hints)
        cls.__cls_store = _utils.Store(
            fields=fields,
            const_hints=types.MappingProxyType(const_hints),
            inst_hints=types.MappingProxyType(inst_hints),
            consts=types.MappingProxyType(consts),
            defaults=frozenset(defaults),
        )
        if Toggles.FORBID_NON_NAMED_FIELDS in cls.__toggles__:
            _utils.check_non_named_fields(cls.__cls_store)
        if Toggles.FORBID_UNANNOTATED_FIELDS in cls.__toggles__:
            _utils.check_unannotated_fields(cls.__cls_store)

    def __init__(self, **kwargs: t.Any) -> None:  # noqa: ANN401
        self.__iter = None
        self.__kwargs = kwargs.copy()
        self.__process_toggles()
        self.__populate_attrs()
        msg = self.__process_template(self.as_dict())
        msg = self._override_message(self.__cls_store, kwargs, msg)
        super().__init__(msg)

    def __process_toggles(self) -> None:
        """Trigger toggles."""
        store = self.__cls_store
        kws = frozenset(self.__kwargs)

        if Toggles.FORBID_MISSING_FIELDS in self.__toggles__:
            _utils.check_missing_fields(store, kws)

        if Toggles.FORBID_UNDECLARED_FIELDS in self.__toggles__:
            _utils.check_undeclared_fields(store, kws)

        if Toggles.FORBID_KWARG_CONSTS in self.__toggles__:
            _utils.check_kwarg_consts(store, kws)

    def __populate_attrs(self) -> None:
        """Set hinted kwargs as exception attributes."""
        for k, v in self.__kwargs.items():
            if k in self.__cls_store.inst_hints:
                setattr(self, k, v)

    def __process_template(self, data: t.Dict[str, t.Any]) -> str:
        """Format the error template from provided data (kwargs & defaults)."""
        kwargs = self.__cls_store.consts.copy()
        kwargs.update(data)
        return _utils.format_template(self.__template__, kwargs)

    def _override_message(  # noqa: PLR6301
        self,
        store: _utils.Store,  # noqa: ARG002
        kwargs: t.Dict[str, t.Any],  # noqa: ARG002
        msg: str,
    ) -> str:
        """
        Adapter method to wedge user logic into izulu machinery.

        This is the place to override message/formatting if regular mechanics
        don't work for you. It has to return original or your flavored message.
        The method is invoked between izulu preparations and original
        ``Exception`` constructor receiving the result of this hook.

        You can also do any other logic here. You will be provided with
        complete set of prepared data from izulu. But it's recommended
        to use classic OOP inheritance for ordinary behaviour extension.

        Args:
            store: dataclass containing inner error class specifications
            kwargs: original kwargs from user
            msg: formatted message from the error template

        """
        return msg

    def __repr__(self) -> str:
        kwargs = _utils.join_kwargs(**self.as_dict())
        return f"{self.__module__}.{self.__class__.__qualname__}({kwargs})"

    def __copy__(self) -> Error:
        return type(self)(**self.as_dict())

    def __deepcopy__(self, memo: t.Dict[int, t.Any]) -> Error:
        id_ = id(self)
        if id_ not in memo:
            kwargs = {
                k: copy.deepcopy(v, memo) for k, v in self.as_dict().items()
            }
            memo[id_] = type(self)(**kwargs)
        return t.cast("Error", memo[id_])

    def __reduce__(self) -> t.Tuple[t.Any, ...]:
        return functools.partial(self.__class__, **self.as_dict()), tuple()

    def as_str(self) -> str:
        """Represent error as exception type with message."""
        return f"{self.__class__.__qualname__}: {self}"

    def as_kwargs(self) -> t.Dict[str, t.Any]:
        """Return the copy of original kwargs used to initialize the error."""
        return self.__kwargs.copy()

    def as_dict(self, *, wide: bool = False) -> t.Dict[str, t.Any]:
        """
        Represent error as dict of fields including default values.

        By default, only *instance* data and defaults are provided.

        Args:
            wide: if ``True`` *class* defaults will be included in result

        """
        d = self.__kwargs.copy()
        for field in self.__cls_store.defaults:
            d.setdefault(field, getattr(self, field))
        if wide:
            for field, const in self.__cls_store.consts.items():
                d.setdefault(field, const)
        return d


@t.overload
def dump(exc: BaseException, /) -> ErrorDumpDict: ...


@t.overload
def dump(
    exc: BaseException,
    /,
    *excs: BaseException,
) -> t.Tuple[ErrorDumpDict, ...]: ...


def dump(
    exc: BaseException,
    /,
    *excs: BaseException,
) -> t.Union[ErrorDumpDict, t.Tuple[ErrorDumpDict, ...]]:
    fields = None
    if isinstance(exc, Error):
        fields = exc.as_dict(wide=True)

    dumped: ErrorDumpDict = dict(
        type=exc.__class__.__name__,
        reason=str(exc),
        fields=fields,
        details={},
    )

    if excs:
        return dumped, *(dump(e) for e in excs)

    return dumped
